/**
 * Copyright (C) 2010-12 Brookhaven National Laboratory
 * All rights reserved. Use is subject to license terms.
 */
package org.epics.pvmanager;

/**
 * An event for the reader.
 * <p>
 * An event can be trigger by:
 * <ul>
 * <li>a connection change; the connection either connected or disconnected</li>
 * <li>an exception; an error either at the datasource or while calculating the expression</li>
 * <li>a value change</li>
 * </ul>
 * One event can have multiple triggers. You can use the mask or the methods to check what condition applies.
 *
 * @param <T> the type of the reader
 * @author carcassi
 */
public class PVReaderEvent<T> {
    
    /**
     * Mask for a connection event.
     */
    public static int CONNECTION_MASK = 0b000001;
    /**
     * Mask for an error event.
     */
    public static int EXCEPTION_MASK  = 0b000010;
    /**
     * Mask for a value change.
     */
    public static int VALUE_MASK      = 0b000100;
    
    private final int notificationMask;
    private final PVReader<T> pvReader;

    PVReaderEvent(int notificationMask, PVReader<T> pvReader) {
        this.notificationMask = notificationMask;
        this.pvReader = pvReader;
    }

    /**
     * The reader that generated the event.
     *
     * @return the pv reader
     */
    public PVReader<T> getPvReader() {
        return pvReader;
    }

    /**
     * The mask for the event.
     * 
     * @return event mask
     */
    public int getNotificationMask() {
        return notificationMask;
    }
    
    /**
     * Whether this event was generated by a connection change.
     *
     * @return true if the connection status changed from the last notification
     */
    public boolean isConnectionChanged() {
        return (notificationMask & CONNECTION_MASK) != 0;
    }
    
    /**
     * Whether this event was generated by a value change.
     *
     * @return true if the value changed from the last notification
     */
    public boolean isValueChanged() {
        return (notificationMask & VALUE_MASK) != 0;
    }
    
    /**
     * Whether this event was generated in response to an error.
     *
     * @return true if a new exception is available
     */
    public boolean isExceptionChanged() {
        return (notificationMask & EXCEPTION_MASK) != 0;
    }

    @Override
    public String toString() {
        StringBuilder builder = new StringBuilder();
        builder.append("[ ");
        boolean first = true;
        if (isConnectionChanged()) {
            builder.append("CONN");
            first = false;
        }
        if (isValueChanged()) {
            if (!first) {
                builder.append(" | ");
            }
            builder.append("VAL");
            first = false;
        }
        if (isExceptionChanged()) {
            if (!first) {
                builder.append(" | ");
            }
            builder.append("EXC");
        }
        builder.append(" ]");
        return builder.toString();
    }
    
}
